@@ -2,42 +2,63 @@ from __future__ import annotations
 
 from abc import ABC, abstractmethod
 from dataclasses import dataclass
-from functools import wraps
-from typing import (
-    Any,
-    Callable,
-    Generic,
-    NoReturn,
-    ParamSpec,
-    TypeGuard,
-    TypeVar,
-    cast,
-    overload,
-)
+from typing import Any, Callable, Generic, NoReturn, TypeGuard, TypeVar, cast, overload
 
 from .panic import Panic
 
 T_co = TypeVar("T_co", covariant=True)
 E_co = TypeVar("E_co", covariant=True)
+T = TypeVar("T")
+E = TypeVar("E")
 U = TypeVar("U")
 R = TypeVar("R")
 F = TypeVar("F")
-P = ParamSpec("P")
 
 
 class ResultShortcutError(Exception, Generic[E_co]):
     def __init__(self, error: Err[E_co]) -> None:
         super().__init__(
-            f"The Q operator used without rustshed.result_shortcut decorator!"
+            "The Q operator used without rustshed.result_shortcut decorator!"
         )
         self.error = error
 
+    def __class_getitem__(cls, item: Any) -> type[ResultShortcutError[E_co]]:
+        # enables applying a type parameter for generic Exception class
+        # except ResultShortcutError[E] as err:
+        return cls
+
 
 class OptionShortcutError(Exception):
     ...
 
 
 class _BaseOption(ABC, Generic[T_co]):
+    """
+    Optional values.
+
+    The Option type represents an optional value: every `Option[T]` is either `Some[T]`
+    and contains a value of type `T`, or `Null` (`None` in Rust), and does not.
+
+    ### Example
+
+    Options are commonly paired with pattern matching
+    to query the presence of a value and take action,
+    always accounting for the Null case.
+
+    ```
+    def divide(numerator: float, denominator: float) -> Option[float]:
+        if denominator == 0.0:
+            return Null
+        return Some(numerator / denominator)
+
+    match divide(2.0, 3.0):
+        case Some(value):
+            print(f"Result: {value}")
+        case Null:
+            print("Cannot divide by 0)
+    ```
+    """
+
     @abstractmethod
     def is_some(self) -> TypeGuard[Some[T_co]]:
         """
@@ -47,7 +68,8 @@ class _BaseOption(ABC, Generic[T_co]):
     @abstractmethod
     def is_some_and(self, f: Callable[[T_co], bool]) -> bool:
         """
-        Returns true if the option is a Some and the value inside of it matches a predicate.
+        Returns true if the option is a Some and the value inside
+        of it matches a predicate.
         """
 
     @abstractmethod
@@ -98,6 +120,12 @@ class _BaseOption(ABC, Generic[T_co]):
         Maps an Option<T> to Option<U> by applying a function to a contained value.
         """
 
+    @abstractmethod
+    def inspect(self, f: Callable[[T_co], Any]) -> Option[T_co]:
+        """
+        Calls the provided closure with a reference to the contained value (if Some).
+        """
+
     @abstractmethod
     def map_or(self, default: U, f: Callable[[T_co], U]) -> U:
         """
@@ -116,6 +144,24 @@ class _BaseOption(ABC, Generic[T_co]):
         or applies a different function to the contained value (if any).
         """
 
+    @abstractmethod
+    def ok_or(self, err: E) -> Result[T_co, E]:
+        """
+        Transforms the Option<T> into a Result<T, E>,
+        mapping Some(v) to Ok(v) and Null to Err(err).
+
+        Arguments passed to ok_or are eagerly evaluated;
+        if you are passing the result of a function call,
+        it is recommended to use ok_or_else, which is lazily evaluated.
+        """
+
+    @abstractmethod
+    def ok_or_else(self, err: Callable[[], E]) -> Result[T_co, E]:
+        """
+        Transforms the Option<T> into a Result<T, E>,
+        mapping Some(v) to Ok(v) and Null to Err(err()).
+        """
+
     @abstractmethod
     def and_(self, optb: Option[U]) -> Option[U]:
         """
@@ -199,6 +245,15 @@ class _BaseOption(ABC, Generic[T_co]):
         Otherwise, (Null, Null) is returned.
         """
 
+    @abstractmethod
+    def transpose(self: _BaseOption[_BaseResult[T, E]]) -> Result[Option[T], E]:
+        """
+        Transposes an Option of a Result into a Result of an Option.
+
+        Null will be mapped to Ok(Null).
+        Some(Ok(T)) and Some(Err(E)) will be mapped to Ok(Some(T)) and Err(E).
+        """
+
     @abstractmethod
     def flatten(self: _BaseOption[_BaseOption[U]]) -> Option[U]:
         """
@@ -241,12 +296,22 @@ class Some(_BaseOption[T_co]):
     def map(self, f: Callable[[T_co], U]) -> Some[U]:
         return Some(f(self.value))
 
+    def inspect(self, f: Callable[[T_co], Any]) -> Option[T_co]:
+        f(self.value)
+        return self
+
     def map_or(self, default: U, f: Callable[[T_co], U]) -> U:
         return f(self.value)
 
     def map_or_else(self, default: Callable[[], U], f: Callable[[T_co], U]) -> U:
         return f(self.value)
 
+    def ok_or(self, err: Any) -> Ok[T_co]:
+        return Ok(self.value)
+
+    def ok_or_else(self, err: Any) -> Ok[T_co]:
+        return Ok(self.value)
+
     def and_(self, optb: Option[U]) -> Option[U]:
         return optb
 
@@ -310,6 +375,28 @@ class Some(_BaseOption[T_co]):
             case _:
                 return Null, Null
 
+    @overload
+    def transpose(self: Some[Ok[T]]) -> Ok[Some[T]]:
+        ...  # pragma: no cover
+
+    @overload
+    def transpose(self: Some[Err[E]]) -> Err[E]:
+        ...  # pragma: no cover
+
+    @overload
+    def transpose(self: Option[Result[T, E]]) -> Result[Option[T], E]:
+        ...  # pragma: no cover
+
+    def transpose(self: Option[Result[T, E]]) -> Result[Option[T], E]:
+        match self:
+            case Some(Ok(value)):
+                return Ok(Some(value))
+            case Some(Err(error)):
+                return Err(error)
+            case _:  # pragma: no cover
+                # it will never happen
+                raise RuntimeError
+
     @overload
     def flatten(self: Some[NullType]) -> NullType:
         ...  # pragma: no cover
@@ -322,7 +409,7 @@ class Some(_BaseOption[T_co]):
     def flatten(self: Some[Option[U]]) -> Option[U]:
         ...  # pragma: no cover
 
-    def flatten(self: Some[Any]) -> Any:
+    def flatten(self: Some[NullType | Some[U]]) -> NullType | Some[U]:
         match self:
             case Some(NullType()):
                 return Null
@@ -330,7 +417,7 @@ class Some(_BaseOption[T_co]):
                 return Some(val)
             case _:  # pragma: no cover
                 # it will never happen
-                raise
+                raise RuntimeError
 
     @property
     def Q(self) -> T_co:
@@ -339,6 +426,9 @@ class Some(_BaseOption[T_co]):
 
 @dataclass
 class NullType(_BaseOption[Any]):
+    def __str__(self) -> str:
+        return "Null"
+
     def is_some(self) -> TypeGuard[Some[Any]]:
         return False
 
@@ -363,19 +453,28 @@ class NullType(_BaseOption[Any]):
     def map(self, f: Callable[[Any], Any]) -> NullType:
         return self
 
+    def inspect(self, f: Callable[[Any], Any]) -> NullType:
+        return self
+
     def map_or(self, default: U, f: Callable[[Any], U]) -> U:
         return default
 
     def map_or_else(self, default: Callable[[], U], f: Callable[[Any], U]) -> U:
         return default()
 
+    def ok_or(self, err: E) -> Result[T_co, E]:
+        return Err(err)
+
+    def ok_or_else(self, err: Callable[[], E]) -> Result[T_co, E]:
+        return Err(err())
+
     def and_(self, optb: Option[Any]) -> NullType:
         return self
 
     def and_then(self, f: Callable[[Any], Option[U]]) -> NullType:
         return self
 
-    def filter(self, predicate: Callable[[T_co], bool]) -> NullType:
+    def filter(self, predicate: Callable[[Any], bool]) -> NullType:
         return self
 
     def or_(self, optb: Option[T_co]) -> Option[T_co]:
@@ -408,6 +507,9 @@ class NullType(_BaseOption[Any]):
     def unzip(self) -> tuple[NullType, NullType]:
         return (Null, Null)
 
+    def transpose(self: _BaseOption[Result[T, E]]) -> Result[NullType, E]:
+        return Ok(Null)
+
     def flatten(self: _BaseOption[_BaseOption[Any]]) -> NullType:
         return Null
 
@@ -421,17 +523,6 @@ Option = Some[T_co] | NullType
 Null = NullType()
 
 
-def to_option(f: Callable[P, T_co]) -> Callable[P, Option[T_co]]:
-    @wraps(f)
-    def wrapper(*args: P.args, **kwargs: P.kwargs) -> Option[T_co]:
-        try:
-            return Some(f(*args, **kwargs))
-        except:
-            return Null
-
-    return wrapper
-
-
 class _BaseResult(ABC, Generic[T_co, E_co]):
     @abstractmethod
     def is_ok(self) -> TypeGuard[Ok[T_co]]:
@@ -499,7 +590,7 @@ class _BaseResult(ABC, Generic[T_co, E_co]):
         """
 
     @abstractmethod
-    def map_or_else(self, default: Callable[[E_co], U], fn: Callable[[T_co], U]) -> U:
+    def map_or_else(self, default: Callable[[E_co], U], f: Callable[[T_co], U]) -> U:
         """
         Maps a Result<T, E> to U by applying fallback function default
         to a contained Err value, or function f to a contained Ok value.
@@ -518,6 +609,18 @@ class _BaseResult(ABC, Generic[T_co, E_co]):
         while handling an error.
         """
 
+    @abstractmethod
+    def inspect(self, f: Callable[[T_co], Any]) -> Result[T_co, E_co]:
+        """
+        Calls the provided closure with a reference to the contained value (if Ok).
+        """
+
+    @abstractmethod
+    def inspect_err(self, f: Callable[[E_co], Any]) -> Result[T_co, E_co]:
+        """
+        Calls the provided closure with a reference to the contained error (if Err).
+        """
+
     @abstractmethod
     def expect(self, msg: str) -> T_co:
         """
@@ -619,10 +722,19 @@ class _BaseResult(ABC, Generic[T_co, E_co]):
         Returns true if the result is an Err value containing the given value.
         """
 
+    @abstractmethod
+    def transpose(self: _BaseResult[_BaseOption[T], E]) -> Option[Result[T, E]]:
+        """
+        Transposes a Result of an Option into an Option of a Result.
+
+        Ok(Null) will be mapped to Null.
+        Ok(Some(T)) and Err(E) will be mapped to Some(Ok(T)) and Some(Err(E)).
+        """
+
     @property
     @abstractmethod
     def Q(self) -> T_co:
-        ...
+        ...  # pragma: no cover
 
 
 @dataclass
@@ -653,12 +765,19 @@ class Ok(_BaseResult[T_co, Any]):
     def map_or(self, default: U, fn: Callable[[T_co], U]) -> U:
         return fn(self.value)
 
-    def map_or_else(self, default: Callable[[Any], U], fn: Callable[[T_co], U]) -> U:
-        return fn(self.value)
+    def map_or_else(self, default: Callable[[object], U], f: Callable[[T_co], U]) -> U:
+        return f(self.value)
 
     def map_err(self, op: Callable[[Any], Any]) -> Ok[T_co]:
         return self
 
+    def inspect(self, f: Callable[[T_co], Any]) -> Ok[T_co]:
+        f(self.value)
+        return self
+
+    def inspect_err(self, f: Callable[[Any], Any]) -> Result[T_co, E_co]:
+        return self
+
     def expect(self, msg: str) -> T_co:
         return self.value
 
@@ -695,6 +814,28 @@ class Ok(_BaseResult[T_co, Any]):
     def contains_err(self, f: Any) -> bool:
         return False
 
+    @overload
+    def transpose(self: Ok[Some[T]]) -> Some[Ok[T]]:
+        ...  # pragma: no cover
+
+    @overload
+    def transpose(self: Ok[NullType]) -> NullType:
+        ...  # pragma: no cover
+
+    @overload
+    def transpose(self: Ok[Option[T]]) -> Option[Ok[T]]:
+        ...  # pragma: no cover
+
+    def transpose(self: Ok[Option[T]]) -> Option[Ok[T]]:
+        match self:
+            case Ok(Some(value)):
+                return Some(Ok(value))
+            case Ok(NullType()):
+                return Null
+            case _:  # pragma: no cover
+                # it will never happen
+                raise RuntimeError
+
     @property
     def Q(self) -> T_co:
         return self.value
@@ -728,12 +869,19 @@ class Err(_BaseResult[Any, E_co]):
     def map_or(self, default: U, fn: Callable[[Any], U]) -> U:
         return default
 
-    def map_or_else(self, default: Callable[[E_co], U], fn: Callable[[Any], U]) -> U:
+    def map_or_else(self, default: Callable[[E_co], U], f: Callable[[Any], U]) -> U:
         return default(self.error)
 
     def map_err(self, op: Callable[[E_co], F]) -> Err[F]:
         return Err(op(self.error))
 
+    def inspect(self, f: Callable[[Any], Any]) -> Err[E_co]:
+        return self
+
+    def inspect_err(self, f: Callable[[E_co], Any]) -> Err[E_co]:
+        f(self.error)
+        return self
+
     def expect(self, msg: str) -> NoReturn:
         raise Panic(msg)
 
@@ -770,60 +918,13 @@ class Err(_BaseResult[Any, E_co]):
     def contains_err(self, f: object) -> bool:
         return self.error == f
 
+    def transpose(self: Err[E]) -> Some[Err[E]]:
+        return Some(self)
+
     @property
     def Q(self) -> NoReturn:
         raise ResultShortcutError(self)
 
 
 Result = Ok[T_co] | Err[E_co]
-
-
-E_rr = TypeVar("E_rr", bound=Exception)
-
-
-class to_result_type:
-    def __call__(self, f: Callable[P, T_co]) -> Callable[P, Result[T_co, Exception]]:
-        @wraps(f)
-        def wrapper(*args: P.args, **kwargs: P.kwargs) -> Result[T_co, Exception]:
-            try:
-                return Ok(f(*args, **kwargs))
-            except Exception as err:
-                return Err(err)
-
-        return wrapper
-
-    def __getitem__(
-        self, _: type[E_rr]
-    ) -> Callable[[Callable[P, T_co]], Callable[P, Result[T_co, E_rr]]]:
-        return self  # type: ignore
-
-
-to_result = to_result_type()
-
 IOResult = Result[T_co, IOError]
-
-RT = TypeVar("RT")
-
-
-def result_shortcut(
-    f: Callable[P, Result[T_co, E_co]]
-) -> Callable[P, Result[T_co, E_co]]:
-    @wraps(f)
-    def wrapper(*args: P.args, **kwargs: P.kwargs) -> Result[T_co, E_co]:
-        try:
-            return f(*args, **kwargs)
-        except ResultShortcutError as err:
-            return cast(Result[T_co, E_co], err.error)
-
-    return wrapper
-
-
-def option_shortcut(f: Callable[P, Option[T_co]]) -> Callable[P, Option[T_co]]:
-    @wraps(f)
-    def wrapper(*args: P.args, **kwargs: P.kwargs) -> Option[T_co]:
-        try:
-            return f(*args, **kwargs)
-        except OptionShortcutError:
-            return cast(Option[T_co], Null)
-
-    return wrapper
