@@ -1,2 +1,2 @@
 class Panic(RuntimeError):
-    ...
+    ...  # pragma: no cover
